#include <mex.h>
#include <string.h>

#define ISREALDOUBLE(P) (!mxIsComplex(P) && mxIsDouble(P))
#define STRINGIZE(x) #x
#define MACROSTRING(x) STRINGIZE(x)
#define MALLOCFAIL mexErrMsgTxt("Memory allocation failure (mex line " MACROSTRING(__LINE__) ")")
typedef struct nodestruct {size_t vert; double prob; struct nodestruct* next; int disttoend;} nodestruct;
#define STRINGINPUTBUFFERSIZE 10

size_t hpp(mwIndex curr);

size_t *nnltp;
bool  *started, *unfinished;
nodestruct **nodelinkstopaths;
mwIndex *ir, *jc, stop;
double *pr, minprob;

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
// 	[paths,probs] = highprobpaths(probmat,minprob,start,stop,outmode);  % outmode is optional but must be 'full' or 'firstlast'
	bool sparsecost;
	double *pathprobs, *pathverts;
	const mxArray *probmat;
	mxArray *mxpv;
	mwIndex start;
	size_t i, j, k, m, n, npath, plen, maxplen;
	nodestruct *startnode, *pathnode;
	char outmodestr[STRINGINPUTBUFFERSIZE];
	enum {cell,array,firstlast} outputmode;
	
	if(nrhs < 4)
		mexErrMsgTxt("Too few inputs");
	if(nrhs > 5)
		mexErrMsgTxt("Too many inputs");
	if(nlhs > 2)
		mexErrMsgTxt("Too many outputs");
	
	if(!ISREALDOUBLE(prhs[0]) || mxGetNumberOfDimensions(prhs[0])>2)
		mexErrMsgTxt("PROBMAT must be a real double matrix");
	for(i = 1; i < (nrhs>4 ? 4 : nrhs); i++)
		if(!ISREALDOUBLE(prhs[i]) || mxGetNumberOfElements(prhs[i]) != 1)
			mexErrMsgTxt("Inputs except PROBMAT must be real double scalars");
	if(nrhs > 4)
	{
		if(mxIsChar(prhs[4]))
		{
			mxGetString(prhs[4],outmodestr,STRINGINPUTBUFFERSIZE);
			if(_strcmpi(outmodestr,"cell") == 0)
				outputmode = cell;
			else if(_strcmpi(outmodestr,"array") == 0)
				outputmode = array;
			else if(_strcmpi(outmodestr,"firstlast") == 0)
				outputmode = firstlast;
			else
				mexErrMsgTxt("OUTMODE must be either 'cell', 'array', or 'firstlast'");
		}
		else
			mexErrMsgTxt("OUTMODE must be either 'cell', 'array', or 'firstlast'");
	}
	else
		outputmode = cell;
	
	probmat = prhs[0];
	minprob = *mxGetPr(prhs[1]);
	if(minprob < 0)
		mexErrMsgTxt("Input MINPROB must greater than or equal to zero");
	start = (mwIndex) *mxGetPr(prhs[2]) - 1; // convert to C zero-indexing
	stop = (mwIndex) *mxGetPr(prhs[3]) - 1; // convert to C zero-indexing
	if(start == stop)
		mexErrMsgTxt("Input STOP must be different than input START");
	
	m = mxGetM(probmat);
	n = mxGetN(probmat);
	if(m < 1 || m != n)
		mexErrMsgTxt("PROBMAT must be square and nonempty");
	if(start < 0 || start >= n || stop <0 || stop >= n)
		mexErrMsgTxt("START and STOP must index into PROBMAT");
	
	sparsecost = mxIsSparse(probmat);
	if(sparsecost)
	{
		ir = mxGetIr(probmat);
		jc = mxGetJc(probmat);
	}
	else // probmat isn't sparse, but we can pretend it is
	{
		mexWarnMsgTxt("Performance is usually better when PROBMAT is sparse"); // user should really use sparse matrices, even if the matrix is somewhat dense
		ir = (mwIndex*) mxMalloc(sizeof(mwIndex)*m*n);
		jc = (mwIndex*) mxMalloc(sizeof(mwIndex)*(n+1));
		if(ir == NULL || jc == NULL)
			MALLOCFAIL;
		k = 0;
		for(j = 0; j < n; j++) // make ir and jc as if probmat were sparse
		{
			jc[j] = k;
			for(i = 0; i < m; i++, k++)
				ir[k] = i;
		}
		jc[j] = k; // jc[n] = m*n;
	}
	pr = mxGetPr(probmat);
	
	nodelinkstopaths = mxMalloc(sizeof(nodestruct*)*n);
	nnltp = mxMalloc(sizeof(size_t)*n); // number of entries for each vertex in nodelinkstopaths
	started = mxMalloc(sizeof(bool)*n); // use to detect cycles
	unfinished = mxMalloc(sizeof(bool)*n); // use to detect cycles and track what we've done
	if(nodelinkstopaths == NULL || nnltp == NULL || started == NULL || unfinished == NULL)
		MALLOCFAIL;
	for(i = 0; i < n; i++)
	{
		started[i] = false; // track what vertices we've started
		unfinished[i] = true; // track what vertices we haven't finished
	}
	
	nodelinkstopaths[stop] = mxMalloc(sizeof(nodestruct));
	if(nodelinkstopaths[stop] == NULL)
		MALLOCFAIL;
	nodelinkstopaths[stop]->prob = 1; // probability starts at 1 for every path
	nodelinkstopaths[stop]->vert = stop;
	nodelinkstopaths[stop]->disttoend = 0;
	nodelinkstopaths[stop]->next = NULL;
	nnltp[stop] = 1;
	started[stop] = true;
	unfinished[stop] = false;
	
	npath = hpp(start);
// 	mexPrintf("found %i paths\n",npath);  mexEvalString("drawnow;");
	
	if(!sparsecost) // free these if we had to generate them
	{
		mxFree(ir);
		mxFree(jc);
	}
	
	switch(outputmode)
	{
		case cell:
			plhs[0] = mxCreateCellMatrix(npath,1);
			startnode = &nodelinkstopaths[start][0]; // get node at start of path
			pathprobs = mxMalloc(sizeof(double)*npath);
			if(npath > 0 && pathprobs == NULL)
				MALLOCFAIL;
			for(i = 0; i < npath; i++)
			{
				plen = startnode->disttoend - 1; // path length
				pathverts = mxMalloc(sizeof(double)*plen);
				if(plen > 0 && pathverts == NULL)
					MALLOCFAIL;
				pathnode = startnode->next;
				pathprobs[i] = startnode->prob;
				for(j = 0; j < plen; j++)
				{
					pathverts[j] = (double) (pathnode->vert + 1); // convert back to MATLAB one-indexing
					pathnode = pathnode->next; // continue down path
				}
				
				mxpv = mxCreateDoubleMatrix(0,1,mxREAL); mxSetM(mxpv,plen); // create unallocated array
				mxSetPr(mxpv,pathverts); // assign data to array
				mxSetCell(plhs[0],i,mxpv); // place array in output cell array
				startnode++; // next path-starting node
			}
			break;
			
		case array:
			maxplen = 0;
			startnode = &nodelinkstopaths[start][0]; // get node at start of path
			for(i = 0; i < npath; i++)
			{
				if(startnode->disttoend > maxplen)
					maxplen = startnode->disttoend;
				startnode++;
			}
			if(maxplen > 0)
				maxplen--; // actually disttoend is too big by one
			plhs[0] = mxCreateDoubleMatrix(npath,maxplen,mxREAL); // make zero-allocated array
			pathverts = mxGetPr(plhs[0]);
			startnode = &nodelinkstopaths[start][0]; // get node at start of path
			pathprobs = mxMalloc(sizeof(double)*npath);
			if(npath > 0 && pathprobs == NULL)
				MALLOCFAIL;
			for(i = 0; i < npath; i++)
			{
				plen = startnode->disttoend - 1; // path length
				pathnode = startnode->next;
				pathprobs[i] = startnode->prob;
				k = i;
				for(j = 0; j < plen; j++)
				{
					pathverts[k] = (double) (pathnode->vert + 1); // convert back to MATLAB one-indexing
					pathnode = pathnode->next; // continue down path
					k+=npath; // unfortunately, we're filling this row-major
				}
				startnode++;
			}
			break;
			
		case firstlast:
			startnode = &nodelinkstopaths[start][0]; // get node at start of path
			pathverts = mxMalloc(sizeof(double)*2*npath);
			pathprobs = mxMalloc(sizeof(double)*npath);
			if(npath > 0 && (pathprobs == NULL || pathverts == NULL))
				MALLOCFAIL;
			for(i = 0; i < npath; i++)
			{
				plen = startnode->disttoend - 1; // path length
				if(plen > 0)
				{
					pathnode = startnode->next;
					pathverts[i] = (double) (pathnode->vert + 1); // get start vertex - convert back to MATLAB one-indexing
					for(j = 0; j < plen-1; j++)
						pathnode = pathnode->next; // continue down path
					pathverts[i+npath] = (double) (pathnode->vert + 1); // get stop vertex - convert back to MATLAB one-indexing
				}
				else // special case for 0-length paths
				{
					pathverts[i] = 0;
					pathverts[i+npath] = 0;
				}
				pathprobs[i] = startnode->prob;
				startnode++; // next path-starting node
			}
			plhs[0] = mxCreateDoubleMatrix(0,2,mxREAL); mxSetM(plhs[0],npath); // create unallocated array
			mxSetPr(plhs[0],pathverts); // assign data to array
			break;
	}

	if(nlhs > 1)
	{
		plhs[1] = mxCreateDoubleMatrix(0,1,mxREAL); mxSetM(plhs[1],npath); // create unallocated array
		mxSetPr(plhs[1],pathprobs); // assign data to array
	}
	else
		mxFree(pathprobs); // caller doesn't ask for this, no need anymore
	
	// free everything that's left
	for(i = 0; i < n; i++)
		if(started[i])
			mxFree(nodelinkstopaths[i]);
	mxFree(nodelinkstopaths);
	mxFree(nnltp);
	mxFree(started);
	mxFree(unfinished);
	
	return;
}

size_t hpp(mwIndex curr)
{
	mwIndex jc1, nzcol, *ircol, nvi;
	double *probcol, nprob;
	size_t i, j, maxnnodesfrom, nncurr;
	nodestruct *currnode, *nextnode;
	
	if(started[curr])
		mexErrMsgTxt("Encountered cycle in PROBMAT - require acyclic graph");
	started[curr] = true;
// 	mexPrintf("starting node %i\n",curr); mexEvalString("drawnow;");
	
	jc1 = jc[curr]; // get index into ir
	nzcol = jc[curr+1] - jc1; // number of nonzeros in the column
	ircol = &ir[jc1]; // get locations of nonzeros
	probcol = &pr[jc1]; // get values of nonzeros
	
	maxnnodesfrom = 0;
	for(i = 0; i < nzcol; i++) // count up max possible number of child paths
	{
		if(probcol[i] <= minprob) // link is too weak to have any viable paths
			continue;
		nvi = ircol[i]; // next vertex
		if(unfinished[nvi]) // need to compute paths for nodes if we haven't yet
			hpp(nvi); // explore next node
		maxnnodesfrom += nnltp[nvi]; // bound maximum number of possible links
	}
	
	nodelinkstopaths[curr] = mxMalloc(sizeof(nodestruct)*maxnnodesfrom); // may overallocate, but not too bad
	currnode = nodelinkstopaths[curr];
	if(maxnnodesfrom > 0 && currnode == NULL)
		MALLOCFAIL;
	nncurr = 0;
	for(i = 0; i < nzcol; i++) // now that we have storage for everything we might encounter, evaluate nodes and add to list of connections
	{
		if(probcol[i] <= minprob) // link is too weak to have any viable paths
			continue;
		nvi = ircol[i]; // next vertex
// 		mexPrintf("adding connection %i/%i - node %i with %i links to end\n",i+1,nzcol,nvi,nnltp[nvi]); mexEvalString("drawnow;");
		for(j = 0; j < nnltp[nvi]; j++)
		{
// 			mexPrintf("adding subconnection %i/%i\n",j+1,nnltp[nvi]); mexEvalString("drawnow;");
			nextnode = &nodelinkstopaths[nvi][j];
			nprob = probcol[i] * nextnode->prob; // probability of current step and remainder to end
			if(nprob > minprob) // only connect if new probability is sufficiently large
			{
				currnode->prob = nprob;
				currnode->vert = curr;
				currnode->disttoend = nextnode->disttoend + 1;
				currnode->next = nextnode;
				nncurr++; // we added another node
				currnode++; // move to next allocated node
			}
		}
	}
	nnltp[curr] = nncurr; // number of valid nodes that make it to the end
	unfinished[curr] = false;
// 	mexPrintf("finished node %i, %i paths to end from here\n",curr,nncurr); mexEvalString("drawnow;");
	return nncurr;
}