function [x,obj,ngrad,nobj,converged] = mlpoisson(A,b,y,x0)
% MLPOISSON     Max-likelihood estimation of nonhomogeneous Poisson process
%  [x,obj,ngrad,nobj,converged] = MLPOISSON(A,b,y,x0) is used to compute
%  Poisson process maximum likelihood parameters under a non-negativity
%  constraint. Specifically, it solves:
%  
%  x = argmin_{z>=0}  b'*z - y'*log(A*z)
%  
%  This is done using a quasi-Newton method that approximates the Hessian
%  of the objective as a diagonal matrix and projects the data to the
%  feasible region after each step. Step size is chosen by a backtracking
%  line search.
%  
%  Parameter "x0" is an optional initialization for the search. The value
%  of the objective at the solution is returned as "obj". The number of
%  gradient and (partial) Hessian evaluations used in reaching the solution
%  is returned as "ngrad" and the number of objective evaluations is
%  returned in "nobj".  Output "converged" is true if the solution is found
%  in the allowed iterations.  A failure to converge results in a warning
%  if output "converged" is not requested.
% 
%  It is possible to use a set of function handles in place of argument
%  "A". This is useful when "A" has special structure that allows it to be
%  applied in a computation- or memory-efficient manner. There are two
%  syntaxes for this use:
%   -Cell array "Acell" should operate identically to:
%     Acell{1} == size(A); % a 2-element vector
%     Acell{2} = @(z) A*z;
%     Acell{3} = @(z) A'*z;
%     Acell{4} = @(z) (A.^2)*z; % the transpose of element-wise square of A
%   -Function handle "Afun" should behave such that:
%     Afun(0,0) == size(A); % a 2-element vector
%     Afun(z,1) == A*z;
%     Afun(z,2) == A'*z;
%     Afun(z,3) == (A.^2)'*z; % the transpose of element-wise square of A
% 
%  It is not necessary that the initialization be feasible (infeasible if
%  any(A*x0<=0)), but this saves computation.


%% parameters

maxiter = 500; % maximum number of allowed iterations
objtol = 1e-3; % absolute error tolerance of objfun(x)
argtol = 1e-4; % absolute error tolerance of argument x
convergenceiters = 3; % number of consecutive steps where tolerances must be satisfied

%% input handling

if isnumeric(A) % easy version, we get a matrix
	[m,n] = size(A);
	if isscalar(y) % make scalar y the proper vector
		y = repmat(y,m,1);
	elseif any(size(y) ~= [m,1])
		error('y must be scalar or Mx1 (A is MxN)')
	end
	findy = y~=0;
	y = y(findy); % so we don't evaluate indices where y==0
	A = A(findy,:); % so we don't evaluate indices where y==0
	A2 = A.*A;
	
	Amult = @(z) A*z;
	Atmult = @(z) (z.'*A).';
	A2tmult = @(z) (z.'*A2).';
elseif iscell(A) % use special multiply functions
	if ~isnumeric(A{1}) || numel(A{1}) ~= 2
		error('A{1} must be a 2-element vector specifying the size of the represented A')
	elseif numel(A) < 4 || ~all(cellfun(@(z) isa(z,'function_handle'),A(2:4)))
		error('A{2:4} must be function handles')
	end
	m = A{1}(1);
	n = A{1}(2);

	if isscalar(y) % make scalar y the proper vector
		y = repmat(y,m,1);
	elseif any(size(y) ~= [m,1])
		error('y must be scalar or Mx1 (A is MxN)')
	end
	findy = y~=0;
	y = y(findy); % so we don't evaluate indices where y==0

	if all(findy) % keep all rows
		Amult = A{2};
		Atmult = A{3};
		A2tmult = A{4};
	else  % remove rows of A with y==0
		sel = sparse(1:numel(findy),findy,true,[numel(findy),m]);
		selt = sel';
		Amult = @(z) sel*A{2}(z);
		Atmult = @(z) A{3}(selt*z);
		A2tmult = @(z) A{4}(selt*z);
	end
elseif isa(A,'function_handle') % use special multiply functions
	sizeA = A(0,0);
	if ~isnumeric(sizeA) || numel(sizeA) ~= 2
		error('A(0,0) must be a 2-element vector specifying the size of the represented A')
	end
	m = sizeA(1);
	n = sizeA(2);

	if isscalar(y) % make scalar y the proper vector
		y = repmat(y,m,1);
	elseif any(size(y) ~= [m,1])
		error('y must be scalar or Mx1 (A is MxN)')
	end
	findy = y~=0;
	y = y(findy); % so we don't evaluate indices where y==0

	if all(findy) % keep all rows
		Amult = @(z) A(z,1);
		Atmult = @(z) A(z,2);
		A2tmult = @(z) A(z,3);
	else  % remove rows of A with y==0
		sel = sparse(1:numel(findy),findy,true,[numel(findy),m]);
		selt = sel';
		Amult = @(x) sel*A(z,1);
		Atmult = @(y) A(selt*z,2);
		A2tmult = @(y) A(selt*z,3);
	end
else
	error('Input A must be a matrix, cell array, or function handle');
end

if isscalar(b) % make scalar b the proper vector
	b = repmat(b,n,1);
elseif any(size(b) ~= [n,1])
	error('b must be scalar or Nx1 (A is MxN)')
end

if ~all(isfinite(b)) || ~all(isfinite(y)) % can't solve under any circumstance
	error('Input b or y contains inf or nan values')
end

if nargin < 4 || isempty(x0)
	x = ones(n,1);
elseif isscalar(x0)
	x = repmat(x0,n,1);
elseif all(size(x0) == [n,1])
	if min(x0) < 0
		warning('Initialization x0 violates non-negativity constraint, setting to max(x0,0)...');
	end
	x = max(x0,0);
else
	error('x0 must be scalar, Nx1 (A is MxN), or uninitialized')
end

%% ensure a feasible initialization

minAx = min(Amult(x)); % is the inialization feasible?
if minAx <= 0 % infeasible initialization - add margin and try again
	As = cell(4,1);
	As{1} = [m,n+1];
	As{2} = @(x) Amult(x(1:n))+x(n+1);
	As{3} = @(y) [Atmult(y);sum(y)];
	As{4} = @(y) [A2tmult(y);sum(y)];
	
	s = -minAx + 1; % determine how much margin we need, plus a bit
% 	dummyb = min(sum(abs(b)),realmax/(1e20*s)); % make the dummy variable very expensive (drive it to zero)
	dummyb = 10*sum(sqrt(y)) + m; % want dummyb > sum(y./(A*x)) at solution
% 	dummyb = 1;
	
	xs0 = [x;s];
	ngrad = 0;
	nobj = 0;
	while 1
		bs = [b;dummyb];
		if ~isfinite(bs'*xs0 - y'*log(As{2}(xs0))) % make sure the objective is valid
			error('Failed to create feasible initialization using dummy variable')
		end
		[xs,obj,ngrad1,nobj1,converged] = mlpoisson(As,bs,y,xs0);
		ngrad = ngrad + ngrad1;
		nobj = nobj + nobj1 + 1;
		if xs(n+1) == 0
			break
		elseif ~converged
			error('Optimization could not converge and could not remove dummy variable');
		end
		dummyb = 1e20*dummyb; % regularize harder!
		xs0 = xs; % warm start
	end
	x = xs(1:n);
	if ~converged && nargout < 5
		warning('Optimization did not converge in the allowed iterations (after infeasible initialization)');
	end
	return
end

%% optimization

objfun = @(x) b'*x - y'*log(Amult(x)); % y'*log(A*x);
obj = objfun(x);
nobj = 1;
if ~isreal(obj) || ~isfinite(obj)
	x(1:n-1) = 0; % try again with near-zero initialization (leaves the dummy parameter intact, if we added it)
	obj = objfun(x);
	nobj = nobj+1;
	if ~isreal(obj) || ~isfinite(obj)
		error('Cannot find a feasible initialization')
	end
end

convergencecount = 0;
converged = false;
for i = 1:maxiter
	oAx = 1./(Amult(x)); % 1./(A*x)
	yoAx = y.*oAx;
	yoAx2 = yoAx.*oAx;
	
	grad = b - Atmult(yoAx); % A'*yoAx
	grad2 = A2tmult(yoAx2) + realmin; % (A.^2)'*yoAx2
	
	dir = -grad./grad2;
	dir(x==0 & dir<0) = 0; % we can't push into constraint, so don't try
	
	dirgrad = dir'*grad;
	[obj,x,ll] = stepsize_armijo(objfun,x,dir,dirgrad,obj);
	nobj = nobj+ll;
	
% 	if dirgrad >= -objtol
% 	if max(abs(dir)) <= argtol
	if dirgrad >= -objtol && max(abs(dir)) <= argtol
		convergencecount = convergencecount+1;
		if convergencecount >= convergenceiters
			converged = true;
			break
		end
	else
		convergencecount = 0;
	end
end
ngrad = i;
if ~converged && nargout < 5
	warning('Optimization did not converge in the allowed iterations');
end
% fprintf('Done in %i gradient and %i objective evaluations with objective %g\n',ngrad,nobj,obj)
end

function [fval,x,lleval] = stepsize_armijo(fun,x0,dir,dirgrad,f0)
% for backtracking line search
if isempty(f0)
	lleval = 1;
	f0 = fun(x0);
else
	lleval = 0;
end
if isnan(dirgrad) || dirgrad > 0
	fval = f0;
	x = x0;
	warning('Direction does not seem to be a subgradient');
	return;
end
step = 1;
x = max(x0 + step*dir,0); % projected step in direction dir
lleval = lleval+1;
fval = fun(x);
while ~isreal(fval) || f0 < fval - step*dirgrad/2 % Armijo condition
	step1 = step/2;
	x1 = max(x0 + step*dir,0); % projected step in direction dir
	lleval = lleval+1;
	fval1 = fun(x1);
	if isreal(fval) && fval1 > fval % we did worse by shrinking stepsize
		break % throw away our attempt and go with our previous step
	end
	step = step1;
	fval = fval1;
	x = x1;
end
end